import { LRLanguage, indentNodeProp, continuedIndent, foldNodeProp, foldInside, syntaxTree, LanguageSupport } from '@codemirror/language';
import { linter } from '@codemirror/lint';
import { parseMixed } from '@lezer/common';
import { parser as parser$1 } from '@lezer/javascript';
import { parser, sinceProp, untilProp } from '@surrealdb/lezer';
import { validate, compareVersions } from 'compare-versions';

const surrealqlLanguage = /*@__PURE__*/LRLanguage.define({
    name: "surrealql",
    parser: /*@__PURE__*/parser.configure({
        props: [
            /*@__PURE__*/indentNodeProp.add({
                Object: /*@__PURE__*/continuedIndent({ except: /^\s*}/ }),
                Array: /*@__PURE__*/continuedIndent({ except: /^\s*]/ }),
            }),
            /*@__PURE__*/foldNodeProp.add({
                "Object Array CombinedResult": foldInside,
            }),
        ],
        wrap: /*@__PURE__*/parseMixed((node) => {
            return node.name === "JavaScriptContent" ? { parser: parser$1 } : null;
        }),
    }),
    languageData: {
        closeBrackets: { brackets: ["[", "{", '"', "'", "("] },
        indentOnInput: /^\s*[\]}]$/,
        commentTokens: { line: "--", block: { open: "/*", close: "*/" } },
    },
});
const scopeMap = /*@__PURE__*/new Map([
    ["permission", "PermissionInput"],
    ["index", "IndexInput"],
    ["combined-results", "CombinedResults"],
    ["syntax", "Syntax"],
]);
/**
 * A linter that checks if the syntax tree uses features that are not available in the given SurrealDB version
 *
 * @param version The SurrealDB version to check against (e.g. "2.0.0")
 */
function surrealqlVersionLinter(version) {
    if (!validate(version)) {
        throw new Error(`Invalid SurrealDB version: ${version}`);
    }
    return linter((view) => {
        const diagnostics = [];
        syntaxTree(view.state)
            .cursor()
            .iterate((node) => {
            if (node.from === node.to) {
                return;
            }
            const sinceVersionProp = node.type.prop(sinceProp);
            const untilVersionProp = node.type.prop(untilProp);
            if (sinceVersionProp && compareVersions(version, sinceVersionProp) < 0) {
                diagnostics.push({
                    from: node.from,
                    to: node.to,
                    severity: "error",
                    message: `This syntax is only available on SurrealDB ${sinceVersionProp} and up`,
                });
            }
            if (untilVersionProp && compareVersions(version, untilVersionProp) >= 0) {
                diagnostics.push({
                    from: node.from,
                    to: node.to,
                    severity: "error",
                    message: `This syntax is only available until SurrealDB ${untilVersionProp}`,
                });
            }
        });
        return diagnostics;
    });
}
/**
 * The CodeMirror extension used to add support for the SurrealQL language
 *
 * @param scope Limit the scope of the highlighting
 */
function surrealql(scope) {
    if (!scope) {
        return new LanguageSupport(surrealqlLanguage);
    }
    const scopeId = scopeMap.get(scope);
    if (!scopeId) {
        throw new Error(`Unknown language scope: ${scope}`);
    }
    return new LanguageSupport(surrealqlLanguage.configure({ top: scopeId }));
}

export { surrealql, surrealqlLanguage, surrealqlVersionLinter };
